using System.Security.Cryptography;
using System.Text;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Entities;
using cuqmbr.TravelGuide.Domain.Enums;
using MediatR;

namespace cuqmbr.TravelGuide.Application.Authenticaion.Commands.Register;

public class RegisterCommandHandler : IRequestHandler<RegisterCommand>
{
    private readonly IReadOnlyCollection<IdentityRole> DefaultRoles =
        new IdentityRole[] { IdentityRole.User };

    private readonly UnitOfWork _unitOfWork;
    private readonly PasswordHasherService _passwordHasher;

    public RegisterCommandHandler(UnitOfWork unitOfWork,
        PasswordHasherService passwordHasher)
    {
        _unitOfWork = unitOfWork;
        _passwordHasher = passwordHasher;
    }

    public async Task Handle(RegisterCommand request,
        CancellationToken cancellationToken)
    {
        var datastoreAccount = await _unitOfWork.AccountRepository
            .GetOneAsync(
                e =>
                    e.Email == request.Email ||
                    e.Username == request.Username,
                cancellationToken);

        if (datastoreAccount != null)
        {
            throw new RegistrationException(
                "User with given email or username already registered.");
        }

        
        var defaultRoleIds = (await _unitOfWork.RoleRepository
            .GetPageAsync(
                r => DefaultRoles.Contains(r.Value),
                1, DefaultRoles.Count, cancellationToken))
            .Items
            .Select(r => r.Id);


        var password = Encoding.UTF8.GetBytes(request.Password);

        var salt = RandomNumberGenerator.GetBytes(128 / 8);
        var hash = await _passwordHasher
            .HashAsync(password, salt, cancellationToken);

        var saltBase64 = Convert.ToBase64String(salt);
        var hashBase64 = Convert.ToBase64String(hash);


        var newAccount = new Account
        {
            Username = request.Username,
            Email = request.Email,
            PasswordHash = hashBase64,
            PasswordSalt = saltBase64,
            AccountRoles = defaultRoleIds.Select(id =>
                new AccountRole()
                {
                    RoleId = id
                })
                .ToArray()
        };


        await _unitOfWork.AccountRepository
            .AddOneAsync(newAccount, cancellationToken);

        await _unitOfWork.SaveAsync(cancellationToken);
        _unitOfWork.Dispose();
    }
}
