using System.IdentityModel.Tokens.Jwt;
using System.Security.Claims;
using System.Text;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using MediatR;
using Microsoft.Extensions.Options;
using Microsoft.IdentityModel.Tokens;

namespace cuqmbr.TravelGuide.Application.Authenticaion.Commands.RenewAccessToken;

public class RenewAccessTokenCommandHandler :
    IRequestHandler<RenewAccessTokenCommand, TokensModel>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly JsonWebTokenConfigurationOptions _jwtConfiguration;

    public RenewAccessTokenCommandHandler(UnitOfWork unitOfWork,
        IOptions<ConfigurationOptions> configurationOptions)
    {
        _unitOfWork = unitOfWork;
        _jwtConfiguration = configurationOptions.Value.JsonWebToken;
    }

    public async Task<TokensModel> Handle(
        RenewAccessTokenCommand request, CancellationToken cancellationToken)
    {
        var refreshToken = (await _unitOfWork.RefreshTokenRepository
            .GetOneAsync(e => e.Value == request.RefreshToken,
                cancellationToken));

        if (refreshToken == null)
        {
            throw new AuthenticationException($"Refresh token was not found.");
        }

        if (!refreshToken.IsActive)
        {
            throw new AuthenticationException("Refresh token is inactive.");
        }

        var account = await _unitOfWork.AccountRepository.GetOneAsync(
            a => a.RefreshTokens.Contains(refreshToken),
            a => a.AccountRoles, cancellationToken);

        var jwtSecurityToken = await CreateJwtAsync(account, cancellationToken);
        var accessToken = new JwtSecurityTokenHandler().WriteToken(jwtSecurityToken);

        return new TokensModel(accessToken, refreshToken.Value);
    }

    private async Task<JwtSecurityToken> CreateJwtAsync(
        Account account, CancellationToken cancellationToken)
    {
        var roleIds = account.AccountRoles.Select(ar => ar.RoleId);

        var roles = (await _unitOfWork.RoleRepository
            .GetPageAsync(
                r => roleIds.Contains(r.Id),
                1, roleIds.Count(), cancellationToken))
            .Items.Select(r => r.Value);

        var roleClaims = new List<Claim>();
        foreach (var role in roles)
        {
            roleClaims.Add(new Claim("roles", role.Name));
        }

        var claims = new List<Claim>()
        {
            new Claim(JwtRegisteredClaimNames.Sub, account.Guid.ToString()),
            new Claim(JwtRegisteredClaimNames.Nickname, account.Username),
            new Claim(JwtRegisteredClaimNames.Email, account.Email)
        }
        .Union(roleClaims);

        var expirationDateTimeUtc = DateTime.UtcNow.Add(
            _jwtConfiguration.AccessTokenValidity);

        var symmetricSecurityKey = new SymmetricSecurityKey(
            Encoding.UTF8.GetBytes(_jwtConfiguration.IssuerSigningKey));
        var signingCredentials = new SigningCredentials(
            symmetricSecurityKey, SecurityAlgorithms.HmacSha256);

        var jwtSecurityToken = new JwtSecurityToken(
            issuer: _jwtConfiguration.Issuer,
            audience: _jwtConfiguration.Audience,
            claims: claims,
            expires: expirationDateTimeUtc,
            signingCredentials: signingCredentials);

        return jwtSecurityToken;
    }
}
