using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using Microsoft.Extensions.Localization;
using cuqmbr.TravelGuide.Domain.Enums;
using System.Security.Cryptography;
using cuqmbr.TravelGuide.Application.Common.Services;
using System.Text;

namespace cuqmbr.TravelGuide.Application.Employees.Commands.AddEmployee;

public class AddEmployeeCommandHandler :
    IRequestHandler<AddEmployeeCommand, EmployeeDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    private readonly IStringLocalizer _localizer;
    private readonly PasswordHasherService _passwordHasher;

    public AddEmployeeCommandHandler(UnitOfWork unitOfWork, IMapper mapper,
        IStringLocalizer localizer, PasswordHasherService passwordHasher)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _localizer = localizer;
        _passwordHasher = passwordHasher;
    }

    public async Task<EmployeeDto> Handle(
        AddEmployeeCommand request,
        CancellationToken cancellationToken)
    {
        var parentEntity = await _unitOfWork.CompanyRepository.GetOneAsync(
            e => e.Guid == request.CompanyGuid, cancellationToken);

        if (parentEntity == null)
        {
            throw new NotFoundException(
                $"Parent entity with Guid: {request.CompanyGuid} not found.");
        }

        var entity = await _unitOfWork.EmployeeRepository.GetOneAsync(
            e =>
                e.FirstName == request.FirstName &&
                e.LastName == request.LastName &&
                e.Patronymic == request.Patronymic &&
                e.Sex == request.Sex &&
                e.BirthDate == request.BirthDate &&
                e.CompanyId == parentEntity.Id,
            cancellationToken);

        if (entity != null)
        {
            throw new DuplicateEntityException();
        }


        // Create new account for employee

        var account = await _unitOfWork.AccountRepository.GetOneAsync(
            e => e.Email == request.Email,
            cancellationToken);

        if (account != null)
        {
            throw new DuplicateEntityException();
        }

        var role = (await _unitOfWork.RoleRepository.GetPageAsync(
                1, IdentityRole.Enumerations.Count(), cancellationToken))
            .Items
            .First(r => r.Value.Equals(IdentityRole.CompanyEmployee));

        var salt = RandomNumberGenerator.GetBytes(128 / 8);
        var hash = await _passwordHasher.HashAsync(
            Encoding.UTF8.GetBytes(request.Password),
            salt, cancellationToken);

        var saltBase64 = Convert.ToBase64String(salt);
        var hashBase64 = Convert.ToBase64String(hash);

        account = new Account()
        {
            Username = request.Username,
            Email = request.Email,
            PasswordHash = hashBase64,
            PasswordSalt = saltBase64,
            AccountRoles = new AccountRole[] { new() { RoleId = role.Id } }
        };

        account = await _unitOfWork.AccountRepository.AddOneAsync(
            account, cancellationToken);


        entity = new Employee()
        {
            FirstName = request.FirstName,
            LastName = request.LastName,
            Patronymic = request.Patronymic,
            Sex = request.Sex,
            BirthDate = request.BirthDate,
            Documents = request.Documents.Select(
                d => new EmployeeDocument()
                {
                    DocumentType = d.DocumentType,
                    Information = d.Information
                })
                .ToArray(),
            Company = parentEntity,
            Account = account
        };

        entity = await _unitOfWork.EmployeeRepository.AddOneAsync(
            entity, cancellationToken);


        await _unitOfWork.SaveAsync(cancellationToken);
        _unitOfWork.Dispose();

        return _mapper.Map<EmployeeDto>(entity);
    }
}
