using MediatR;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using cuqmbr.TravelGuide.Application.Common.Services;
using System.Security.Cryptography;
using System.Text;

namespace cuqmbr.TravelGuide.Application
    .Identity.Accounts.Commands.AddAccount;

public class AddAccountCommandHandler : 
    IRequestHandler<AddAccountCommand, AccountDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    private readonly PasswordHasherService _passwordHasher;

    public AddAccountCommandHandler(UnitOfWork unitOfWork,
        IMapper mapper, PasswordHasherService passwordHasher)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _passwordHasher = passwordHasher;
    }

    public async Task<AccountDto> Handle(
        AddAccountCommand request,
        CancellationToken cancellationToken)
    {
        var account = await _unitOfWork.AccountRepository.GetOneAsync(
            e => e.Email == request.Email,
            cancellationToken);

        if (account != null)
        {
            throw new DuplicateEntityException();
        }


        var roles = (await _unitOfWork.RoleRepository
            .GetPageAsync(
                e => request.Roles.Contains(e.Value),
                1, request.Roles.Count, cancellationToken))
            .Items;

        var salt = RandomNumberGenerator.GetBytes(128 / 8);
        var hash = await _passwordHasher.HashAsync(
            Encoding.UTF8.GetBytes(request.Password),
            salt, cancellationToken);

        var saltBase64 = Convert.ToBase64String(salt);
        var hashBase64 = Convert.ToBase64String(hash);

        account = new Account()
        {
            Username = request.Username,
            Email = request.Email,
            PasswordHash = hashBase64,
            PasswordSalt = saltBase64,
            AccountRoles = roles.Select(r => new AccountRole()
            {
                RoleId = r.Id
            })
            .ToArray()
        };

        account = await _unitOfWork.AccountRepository.AddOneAsync(
            account, cancellationToken);

        await _unitOfWork.SaveAsync(cancellationToken);
        _unitOfWork.Dispose();

        return _mapper.Map<AccountDto>(account);
    }
}
