using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.Extensions;
using cuqmbr.TravelGuide.Domain.Entities;

namespace cuqmbr.TravelGuide.Application.Identity.Accounts.Queries.GetAccountsPage;

public class GetAccountsPageQueryHandler :
    IRequestHandler<GetAccountsPageQuery, PaginatedList<AccountDto>>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    public GetAccountsPageQueryHandler(
        UnitOfWork unitOfWork,
        IMapper mapper)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
    }

    public async Task<PaginatedList<AccountDto>> Handle(
        GetAccountsPageQuery request,
        CancellationToken cancellationToken)
    {
        var paginatedList = await _unitOfWork.AccountRepository.GetPageAsync(
            a =>
                (a.Username.ToLower().Contains(request.Search.ToLower()) ||
                a.Email.ToLower().Contains(request.Search.ToLower())) &&
                (request.Roles != null
                    ? request.Roles.All(r => a.AccountRoles.Any(ar => ar.Role.Value == r))
                    : true),
            a => a.AccountRoles,
            request.PageNumber, request.PageSize, cancellationToken);


        var accounts = paginatedList.Items;

        var accountsRoleIds = accounts
            .SelectMany(a => a.AccountRoles)
            .Select(ar => ar.RoleId)
            .Distinct();

        var roles = (await _unitOfWork.RoleRepository
            .GetPageAsync(
                r => accountsRoleIds.Contains(r.Id),
                1, accountsRoleIds.Count(), cancellationToken))
            .Items;

        foreach (var account in accounts)
        {
            account.AccountRoles = account.AccountRoles.Select(ar =>
                new AccountRole()
                {
                    RoleId = ar.RoleId,
                    Role = roles.Single(r => r.Id == ar.RoleId),
                    AccountId = account.Id,
                    Account = account
                })
                .ToArray();
        }


        var mappedItems = _mapper
            .ProjectTo<AccountDto>(accounts.AsQueryable());

        mappedItems = QueryableExtension<AccountDto>
            .ApplySort(mappedItems, request.Sort);

        _unitOfWork.Dispose();

        return new PaginatedList<AccountDto>(
            mappedItems.ToList(),
            paginatedList.TotalCount, request.PageNumber,
            request.PageSize);

        throw new NotImplementedException();
    }
}
