using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using System.Text;
using Newtonsoft.Json;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application.Payments.LiqPay
    .TicketGroups.Commands.ProcessCallback;

public class ProcessCallbackCommandHandler :
    IRequestHandler<ProcessCallbackCommand>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly LiqPayPaymentService _liqPayPaymentService;

    public ProcessCallbackCommandHandler(
        UnitOfWork unitOfWork,
        LiqPayPaymentService liqPayPaymentService)
    {
        _unitOfWork = unitOfWork;
        _liqPayPaymentService = liqPayPaymentService;
    }

    public async Task Handle(
        ProcessCallbackCommand request,
        CancellationToken cancellationToken)
    {
        var isSignatureValid = _liqPayPaymentService
            .IsValidSignature(request.Data, request.Signature);

        if (!isSignatureValid)
        {
            throw new ForbiddenException();
        }

        var dataBytes = Convert.FromBase64String(request.Data);
        var dataJson = Encoding.UTF8.GetString(dataBytes);

        var data = JsonConvert.DeserializeObject<dynamic>(dataJson);

        string status = data.status;

        var ticketGroupGuid = Guid.Parse((string)data.order_id);
        var ticketGroup = await _unitOfWork.TicketGroupRepository
            .GetOneAsync(e => e.Guid == ticketGroupGuid, cancellationToken);

        if (ticketGroup == null ||
            ticketGroup.Status == TicketStatus.Purchased)
        {
            throw new ForbiddenException();
        }

        if (status.Equals("error") || status.Equals("failure"))
        {
            await _unitOfWork.TicketGroupRepository
                .DeleteOneAsync(ticketGroup, cancellationToken);
        }
        else if (status.Equals("success"))
        {
            ticketGroup.Status = TicketStatus.Purchased;
            await _unitOfWork.TicketGroupRepository
                .UpdateOneAsync(ticketGroup, cancellationToken);
        }

        await _unitOfWork.SaveAsync(cancellationToken);
        _unitOfWork.Dispose();
    }
}
