using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using FluentValidation.Results;
using Microsoft.Extensions.Localization;

namespace cuqmbr.TravelGuide.Application.VehicleEnrollments
    .Commands.AddVehicleEnrollment;

public class AddVehicleEnrollmentCommandHandler : 
    IRequestHandler<AddVehicleEnrollmentCommand, VehicleEnrollmentDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    private readonly IStringLocalizer _localizer;

    public AddVehicleEnrollmentCommandHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        IStringLocalizer localizer)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _localizer = localizer;
    }

    public async Task<VehicleEnrollmentDto> Handle(
        AddVehicleEnrollmentCommand request,
        CancellationToken cancellationToken)
    {
        // Check if the vehicle exists.

        var vehicle = await _unitOfWork.VehicleRepository.GetOneAsync(
            e => e.Guid == request.VehicleGuid, cancellationToken);

        if (vehicle == null)
        {
            throw new NotFoundException(
                $"Vehicle with Guid: {request.VehicleGuid} not found.");
        }


        // Check if the route exists.

        var route = await _unitOfWork.RouteRepository.GetOneAsync(
            e => e.Guid == request.RouteGuid, e => e.RouteAddresses,
            cancellationToken);

        if (route == null)
        {
            throw new NotFoundException(
                $"Route with Guid: {request.RouteGuid} not found.");
        }


        // Check if specified vehicle and route compatible.

        if (vehicle.VehicleType != route.VehicleType)
        {
            throw new ValidationException(
                new List<ValidationFailure>
                {
                    new()
                    {
                        PropertyName = nameof(request.VehicleGuid),
                        ErrorMessage = _localizer["Validation.SameVehicleType"]
                    }
                });
        }


        // Check that request has the same Route Addresses
        // as a route data from datastore.

        var sameRouteAddresses = route.RouteAddresses.All(
            ra => request.RouteAddressDetails.Any(
                rad => rad.RouteAddressGuid == ra.Guid));

        if (!sameRouteAddresses)
        {
            throw new NotFoundException(
                $"Not all route addresses are found in a datastore.");
        }

        // Check vehicle enrollments that might overlap with new one.

        var requestDepartureTime = request.DepartureTime;

        var requestTravelTime =
            request.RouteAddressDetails.Aggregate(
                TimeSpan.Zero, (sum, rad) => sum +
                    rad.TimeToNextAddress + rad.CurrentAddressStopTime);

        var requestArrivalTime = requestDepartureTime + requestTravelTime;

        var enrollmentHistory =
            await _unitOfWork.VehicleEnrollmentRepository.GetPageAsync(
                e => 
                    e.Vehicle.Guid == request.VehicleGuid &&
                    e.DepartureTime >= DateTimeOffset.UtcNow.AddDays(-7),
                e => e.RouteAddressDetails,
                1, 200, cancellationToken);

        // Three cases are included:
        //
        // ---SD---------RD----------SA--->
        //                               time
        //
        // ---SD---------RA----------SA--->
        //                               time
        // 
        // ---RD-----SD-------SA-----RA--->
        //                               time
        // Where:
        //  RD - request enrollment departure time
        //  RA - request enrollment arrival time
        //  SD - datastore enrollment (S for store) departure time
        //  SA - datastore enrollment (S for store) arrival time

        var overlappingWithOtherEnrollments = enrollmentHistory.Items
            .Where(ve =>
            {
                var departureTime = ve.DepartureTime;

                var arrivalTime =
                    ve.DepartureTime +
                    ve.RouteAddressDetails
                        .Aggregate(
                            TimeSpan.Zero,
                            (sum, rad) => sum +
                                rad.TimeToNextAddress +
                                rad.CurrentAddressStopTime);

                return
                    (requestDepartureTime >= departureTime &&
                        requestDepartureTime <= arrivalTime) ||
                    (requestArrivalTime >= departureTime &&
                        requestArrivalTime <= arrivalTime) ||
                    (requestDepartureTime <= departureTime &&
                        requestArrivalTime >= arrivalTime);
            })
            .Any();

        if (overlappingWithOtherEnrollments)
        {
            throw new ValidationException(
                new List<ValidationFailure>
                {
                    new()
                    {
                        PropertyName = nameof(request.DepartureTime),
                        ErrorMessage = _localizer["Validation." + 
                            "VehicleEnrollments.OverlapWithOther"]
                    }
                });
        }


        // Check if employee guids are valid

        var employees = (await _unitOfWork.EmployeeRepository
            .GetPageAsync(
                e => request.EmployeeGuids.Contains(e.Guid),
                1, request.EmployeeGuids.Count, cancellationToken))
            .Items;

        if (employees.Count < request.EmployeeGuids.Count)
        {
            throw new NotFoundException();
        }


        // Create entity and add to datastore.

        var entity = new VehicleEnrollment()
        {
            DepartureTime = request.DepartureTime,
            Currency = request.Currency,
            VehicleId = vehicle.Id,
            RouteId = route.Id,
            Route = route,
            Vehicle = vehicle,
            RouteAddressDetails = route.RouteAddresses
                .OrderBy(ra => ra.Order)
                .Select(ra => new RouteAddressDetail()
                {
                    TimeToNextAddress = request.RouteAddressDetails
                        .First(rad => rad.RouteAddressGuid == ra.Guid)
                        .TimeToNextAddress,
                    CostToNextAddress = request.RouteAddressDetails
                        .First(rad => rad.RouteAddressGuid == ra.Guid)
                        .CostToNextAddress,
                    CurrentAddressStopTime = request.RouteAddressDetails
                        .First(rad => rad.RouteAddressGuid == ra.Guid)
                        .CurrentAddressStopTime,
                    RouteAddressId = ra.Id
                })
                .ToArray(),
            VehicleEnrollmentEmployees = request.EmployeeGuids.Select(g => 
                new VehicleEnrollmentEmployee()
                {
                    EmployeeId = employees.Single(e => e.Guid == g).Id
                })
                .ToArray()
        };

        entity = await _unitOfWork.VehicleEnrollmentRepository.AddOneAsync(
            entity, cancellationToken);

        await _unitOfWork.SaveAsync(cancellationToken);

        // Hydrate vehicle enrollment with address information

        var routeAddresses = await _unitOfWork.RouteAddressRepository
            .GetPageAsync(
                e =>
                    request.RouteAddressDetails
                    .Select(rad => rad.RouteAddressGuid)
                    .Contains(e.Guid),
                e => e.Address.City.Region.Country,
                1, entity.RouteAddressDetails.Count(),
                cancellationToken);

        foreach (var rad in entity.RouteAddressDetails)
        {
            rad.RouteAddress = routeAddresses.Items
                .First(ra => ra.Id == rad.RouteAddressId);
        }

        _unitOfWork.Dispose();

        return _mapper.Map<VehicleEnrollmentDto>(entity);
    }
}
