using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application.VehicleEnrollments
    .Queries.GetVehicleEnrollment;

public class GetVehicleEnrollmentQueryHandler :
    IRequestHandler<GetVehicleEnrollmentQuery, VehicleEnrollmentDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly CurrencyConverterService _currencyConverterService;

    public GetVehicleEnrollmentQueryHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        SessionCurrencyService sessionCurrencyService,
        CurrencyConverterService currencyConverterService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _sessionCurrencyService = sessionCurrencyService;
        _currencyConverterService = currencyConverterService;
    }

    public async Task<VehicleEnrollmentDto> Handle(
        GetVehicleEnrollmentQuery request,
        CancellationToken cancellationToken)
    {
        var entity = await _unitOfWork.VehicleEnrollmentRepository.GetOneAsync(
            e => e.Guid == request.Guid, e => e.RouteAddressDetails,
            cancellationToken);

        if (entity == null)
        {
            throw new NotFoundException();
        }

        // Hydrate vehicle enrollment with address,
        // vehicle, route and employee information.

        var routeAddresses = (await _unitOfWork.RouteAddressRepository
            .GetPageAsync(
                e =>
                    entity.RouteAddressDetails
                    .Select(rad => rad.RouteAddressId)
                    .Contains(e.Id),
                e => e.Address.City.Region.Country,
                1, entity.RouteAddressDetails.Count(), cancellationToken))
            .Items;

        foreach (var rad in entity.RouteAddressDetails)
        {
            rad.RouteAddress = routeAddresses
                .First(ra => ra.Id == rad.RouteAddressId);
        }

        var vehicle = await _unitOfWork.VehicleRepository
            .GetOneAsync(e => e.Id == entity.VehicleId, cancellationToken);

        var route = await _unitOfWork.RouteRepository
            .GetOneAsync(e => e.Id == entity.RouteId, cancellationToken);

        var vehicleEnrollmentEmployees = 
            (await _unitOfWork.VehicleEnrollmentEmployeeRepository
                .GetPageAsync(
                    e => e.VehicleEnrollmentId == entity.Id,
                    e => e.Employee,
                    1, int.MaxValue, cancellationToken))
                .Items;

        entity.Vehicle = vehicle;
        entity.Route = route;
        entity.VehicleEnrollmentEmployees =
            vehicleEnrollmentEmployees.ToArray();


        // Convert currency

        // TODO: Replace with AutoMapper Resolver

        if (!_sessionCurrencyService.Currency.Equals(Currency.Default))
        {
            foreach (var rad in entity.RouteAddressDetails)
            {
                rad.CostToNextAddress = await _currencyConverterService
                    .ConvertAsync(rad.CostToNextAddress, entity.Currency,
                        _sessionCurrencyService.Currency, cancellationToken);
            }
            entity.Currency = _sessionCurrencyService.Currency;
        }


        _unitOfWork.Dispose();

        return _mapper.Map<VehicleEnrollmentDto>(entity);
    }
}
