using Microsoft.AspNetCore.Mvc;
using Swashbuckle.AspNetCore.Annotations;
using cuqmbr.TravelGuide.Domain.Enums;
using cuqmbr.TravelGuide.Application.Common.ViewModels;
using cuqmbr.TravelGuide.Application.VehicleEnrollmentSearch;
using cuqmbr.TravelGuide.Application.VehicleEnrollmentSearch
    .Queries.SearchShortest;
using cuqmbr.TravelGuide.Application.VehicleEnrollmentSearch
    .Queries.SearchAll;
using cuqmbr.TravelGuide.Application.VehicleEnrollmentSearch.ViewModels;

namespace cuqmbr.TravelGuide.HttpApi.Controllers;

[Route("vehicleEnrollmentSearch")]
public class VehicleEnrollmentSearchController : ControllerBase
{
    [HttpGet("shortest")]
    [SwaggerOperation("Search shortest vehicle enrollments with transfers")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Search successful",
        typeof(VehicleEnrollmentSearchDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "No enrollments found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<ActionResult<VehicleEnrollmentSearchDto>> SearchShortest(
        [FromQuery] SearchShortestViewModel viewModel,
        CancellationToken cancellationToken)
    {
        return StatusCode(
            StatusCodes.Status201Created,
            await Mediator.Send(
                new SearchShortestQuery()
                {
                    DepartureAddressGuid = viewModel.DepartureAddressUuid,
                    ArrivalAddressGuid = viewModel.ArrivalAddressUuid,
                    DepartureDate = viewModel.DepartureDate,
                    VehicleTypes = viewModel.VehicleTypes
                        .Select(e => VehicleType.FromName(e)).ToHashSet(),
                    ShortestByCost = viewModel.ShortestByCost,
                    ShortestByTime = viewModel.ShortestByTime
                },
                cancellationToken));
    }

    [HttpGet("all")]
    [SwaggerOperation("Search all vehicle enrollments with transfers")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Search successful",
        typeof(IEnumerable<VehicleEnrollmentSearchDto>))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "No enrollments found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<ActionResult<IEnumerable<VehicleEnrollmentSearchDto>>>
        SearchAll(
            [FromQuery] SearchAllViewModel viewModel,
            [FromQuery] SortQuery sortQuery, 
            CancellationToken cancellationToken)
    {
        return StatusCode(
            StatusCodes.Status201Created,
            await Mediator.Send(
                new SearchAllQuery()
                {
                    DepartureAddressGuid = viewModel.DepartureAddressUuid,
                    ArrivalAddressGuid = viewModel.ArrivalAddressUuid,
                    DepartureDate = viewModel.DepartureDate,
                    VehicleTypes = viewModel.VehicleTypes
                        .Select(e => VehicleType.FromName(e)).ToHashSet(),
                    Sort = sortQuery.Sort,
                    TravelTimeGreaterThanOrEqualTo =
                        viewModel.TravelTimeGreaterThanOrEqualTo,
                    TravelTimeLessThanOrEqualTo =
                        viewModel.TravelTimeLessThanOrEqualTo,
                    CostGreaterThanOrEqualTo =
                        viewModel.CostGreaterThanOrEqualTo,
                    CostLessThanOrEqualTo = 
                        viewModel.CostLessThanOrEqualTo,
                    NumberOfTransfersGreaterThanOrEqualTo = 
                        viewModel.NumberOfTransfersGreaterThanOrEqualTo,
                    NumberOfTransfersLessThanOrEqualTo = 
                        viewModel.NumberOfTransfersLessThanOrEqualTo,
                    DepartureTimeGreaterThanOrEqualTo = 
                        viewModel.DepartureTimeGreaterThanOrEqualTo,
                    DepartureTimeLessThanOrEqualTo = 
                        viewModel.DepartureTimeLessThanOrEqualTo,
                    ArrivalTimeGreaterThanOrEqualTo = 
                        viewModel.ArrivalTimeGreaterThanOrEqualTo,
                    ArrivalTimeLessThanOrEqualTo = 
                        viewModel.ArrivalTimeLessThanOrEqualTo
                },
                cancellationToken));
    }
}
