using System.Globalization;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;
using Newtonsoft.Json;

namespace cuqmbr.TravelGuide.Infrastructure.Services;

// https://github.com/fawazahmed0/exchange-api

public sealed class ExchangeApiCurrencyConverterService :
    CurrencyConverterService
{
    private readonly 
        IDictionary<
            DateOnly, IDictionary<
                Currency, IDictionary<
                    Currency, decimal>>> _cache;

    private const string urlFormat = "https://cdn.jsdelivr.net/" +
        "npm/@fawazahmed0/currency-api@{0}/v1/currencies/{1}.json";

    private readonly IHttpClientFactory _httpClientFactory;

    public ExchangeApiCurrencyConverterService(
        IHttpClientFactory httpClientFactory)
    {
        _httpClientFactory = httpClientFactory;

        _cache = 
            new Dictionary<
                DateOnly, IDictionary<
                    Currency, IDictionary<
                        Currency, decimal>>>();
    }

    public async Task<decimal> ConvertAsync(decimal amount, Currency from,
        Currency to, CancellationToken cancellationToken)
    {
        return await ConvertAsync(amount, from, to,
            DateTimeOffset.UtcNow, cancellationToken);
    }

    public async Task<decimal> ConvertAsync(decimal amount, Currency from,
        Currency to, DateTimeOffset time, CancellationToken cancellationToken)
    {
        if (from.Equals(to) || to.Equals(Currency.Default))
        {
            return amount;
        }

        var requestDate = DateOnly.FromDateTime(time.ToUniversalTime().Date);

        // Return cached value if available
        if (_cache.Keys.Contains(requestDate) &&
            _cache[requestDate].Keys.Contains(from) &&
            _cache[requestDate][from].Keys.Contains(to))
        {
            return amount * _cache[requestDate][from][to];
        }


        // Retreive new value from api

        var httpClient = _httpClientFactory.CreateClient();

        var requestDateString = 
            requestDate == DateOnly.FromDateTime(DateTime.UtcNow) ?
            "latest" :
            requestDate.ToString("yyyy-MM-dd", CultureInfo.InvariantCulture);

        var url = String.Format(urlFormat,
            requestDateString, from.Name.ToLower());

        var jsonString = 
            await httpClient.GetStringAsync(url, cancellationToken);

        var obj = JsonConvert.DeserializeObject<dynamic>(jsonString);

        decimal rate = obj[from.Name.ToLower()][to.Name.ToLower()];


        // Cache new value

        if (!_cache.Keys.Contains(requestDate))
        {
            _cache.Add(requestDate,
                new Dictionary<Currency, IDictionary<Currency, decimal>>());
        }

        if (!_cache[requestDate].Keys.Contains(from))
        {
            _cache[requestDate].Add(from, new Dictionary<Currency, decimal>());
        }

        if (!_cache[requestDate][from].Keys.Contains(to))
        {
            _cache[requestDate][from].Add(to, rate);
        }


        return amount * rate;
    }
}
