using System.Security.Cryptography;
using System.Text;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Entities;
using cuqmbr.TravelGuide.Domain.Enums;
using Microsoft.Extensions.DependencyInjection;

namespace cuqmbr.TravelGuide.Persistence;

public static class DbSeeder
{
    public static void Seed(IServiceProvider serviceProvider)
    {
        var unitOfWork =
            serviceProvider.GetRequiredService<UnitOfWork>();

        var passwordHasher =
            serviceProvider.GetRequiredService<PasswordHasherService>();


        // Seed Roles
        {
            var datastoreRoles = unitOfWork.RoleRepository
                .GetPageAsync(1, IdentityRole.Enumerations.Count,
                    CancellationToken.None)
                .Result.Items.Select(r => r.Value);
            
            var roles = IdentityRole.Enumerations.Select(r => r.Value);

            foreach (var role in roles)
            {

                if (datastoreRoles.Contains(role))
                {
                    continue;
                }

                unitOfWork.RoleRepository.AddOneAsync(
                    new Role() { Value = role },
                    CancellationToken.None).Wait();
            }

            unitOfWork.SaveAsync(CancellationToken.None).Wait();
        }

        // Seed Accounts
        {
            var accounts =
                new (string Username, string Email,
                    string Password, IdentityRole[] Roles)[]
                {
                ("admin", "admin", "admin",
                new [] { IdentityRole.Administrator }),
                };

            var roles = unitOfWork.RoleRepository
                .GetPageAsync(1, IdentityRole.Enumerations.Count,
                    CancellationToken.None)
                .Result.Items;

            foreach (var account in accounts)
            {
                var datastoreAccount =
                    unitOfWork.AccountRepository.GetOneAsync(
                        e => e.Email == account.Email, CancellationToken.None)
                    .Result;

                if (datastoreAccount != null)
                {
                    continue;
                }

                var password = Encoding.UTF8.GetBytes(account.Password);

                var salt = RandomNumberGenerator.GetBytes(128 / 8);
                var hash = passwordHasher
                    .HashAsync(password, salt, CancellationToken.None)
                    .Result;

                var saltBase64 = Convert.ToBase64String(salt);
                var hashBase64 = Convert.ToBase64String(hash);

                unitOfWork.AccountRepository.AddOneAsync(
                    new Account()
                    {
                        Username = account.Username,
                        Email = account.Email,
                        PasswordHash = hashBase64,
                        PasswordSalt = saltBase64,
                        AccountRoles = account.Roles.Select(ar =>
                            new AccountRole()
                            {
                                RoleId = roles.Single(dr => dr.Value.Equals(ar)).Id
                            })
                            .ToArray()
                    },
                    CancellationToken.None).Wait();
            }

            unitOfWork.SaveAsync(CancellationToken.None).Wait();
        }

        unitOfWork.Dispose();
    }
}
